public class BinTree<T extends Comparable<T>> {
  T content;
  BinTree<T> left;
  BinTree<T> right;
  BinTree<T> parent;

  public BinTree(T content) {
    this.content = content;
  }

  public boolean contains(T entry) {
    return (this.search(entry) != null);
  }

  public T search(T entry) {
    int comp = this.content.compareTo(entry);
    if (comp == 0) {
      return this.content;
    } else if (comp < 0) {
      if (this.right == null)
        return null;
      return this.right.search(entry);
    } else {
      if (this.left == null)
        return null;
      return this.left.search(entry);
    }
  }

  public void insert(T newContent) {
    int comp = this.content.compareTo(newContent);
    if (comp == 0) {
      return; //element already contained
    } else if (comp < 0) {
      if (this.right == null) {
        // insert here
        this.right = new BinTree<T>(newContent);
        this.right.parent = this;
      } else {
        this.right.insert(newContent);
      }
    } else {
      if (this.left == null) {
        // insert here
        this.left = new BinTree<T>(newContent);
        this.left.parent = this;
      } else {
        this.left.insert(newContent);
      }
    }
  }

  public void delete(T entry) {
    int comp = this.content.compareTo(entry);
    if (comp == 0) {
      //actual bug: throws NullPointer if tree becomes empty

      //do the actual deletion
      if (this.right == null && this.left == null) { //this is a leaf
        if (this.parent.right == this) {
          this.parent.right = null;
        } else if (this.parent.left == this) {
          this.parent.left = null;
        }
      } if (this.right == null) { //only right subtree is empty
        //"pull up" the left subtree
        if (this.parent.right == this) {
          this.parent.right = this.left;
          this.left.parent = this.parent.right;
        } else if (this.parent.left == this) {
          this.parent.left = this.right;
          this.right.parent = this.parent.left;
        }
        return; //for clarity
      } else {
        //possible bug: BinTree<T> nextBigger = this.right;
        BinTree<T> nextBigger = findNextBigger(this);

        // we have to:
        // 1. remove nextBigger from its position, 2. put its right child in its place

        //1. put the right child in nextbigger's place
        if (nextBigger.parent.right == nextBigger) { //possible bug: die abfrage hier vergessen, von immer linkes kind ausgehen
          //this means that nextBigger is the right child of this
          this.right = nextBigger.right;
          if (nextBigger.right != null)
            nextBigger.right.parent = this;
        } else if (nextBigger.parent.left == nextBigger) {
          nextBigger.parent.left = nextBigger.right;
          if (nextBigger.right != null)
            nextBigger.right.parent = nextBigger.parent.left;
        }

        //2. put nextBigger in this's place
        if (this.left != null)
          this.left.parent = nextBigger;
        nextBigger.left = this.left;
        if (this.right != null)
          this.right.parent = nextBigger;
        nextBigger.right = this.right;

        if (this.parent.right == this) {
          this.parent.right = nextBigger;
          nextBigger.parent = this.parent.right;
        } else if (this.parent.left == this) {
          this.parent.left = nextBigger;
          nextBigger.parent = this.parent.left;
        }
      }
    } else if (comp < 0) {
      if (this.right == null) {
        return; //element not found
      } else {
        this.right.delete(entry);
      }
    } else {
      if (this.left == null) {
        return; //element not found
      } else {
        this.left.delete(entry);
      }
    }
 
  }
  
  public BinTree<T> findNextBigger(BinTree<T> start) {
    BinTree<T> nextBigger = start.right;
    while (nextBigger.left != null)
      nextBigger = nextBigger.left;
    return nextBigger;
  }
}
