import java.util.ArrayList;
import java.util.Stack;

public class BinTreeTest {
  public static void main(String[] args) {
    System.out.println(" --- overall test results --- ");
    System.out.println("insertTest result: " + insertTest());
    System.out.println("systemTest result: " + systemTest());
  }
  

  static boolean insertTest() {
    boolean success = true;
    ArrayList<BTInsertTest> insertTests = new ArrayList<>();
    insertTests.add(new BTInsertTest(
      new BinTree<>(new BTEntry("Aeneam", "bla")),
      new BTEntry("ipsum", "bla"),
      "Aeneam,ipsum"));


    BinTree<BTEntry> tree = new BinTree<>(new BTEntry("Aeneam", "bla"));
    tree.insert(new BTEntry("Lorem", "bla"));
    tree.insert(new BTEntry("sit", "bla"));
    tree.insert(new BTEntry("amet", "bla"));
    tree.insert(new BTEntry("Lorem", "bla"));
    insertTests.add(new BTInsertTest(
      tree,
      new BTEntry("ipsum", "bla"),
      "Aeneam,Lorem,amet,ipsum,sit"));

    System.out.println("testing insert");
    for (BTInsertTest it : insertTests) {
      BinTree<BTEntry> bt = it.treeBefore;
      bt.insert(it.element);
      System.out.println("inserting " + it.element 
        + "\nresult:          " + treeToString(bt)
        + "\nexpected result: " + it.treeAfterAsString);
      success &= treeToString(bt).equals(it.treeAfterAsString);
    }
    return success;
  }

  static String treeToString(BinTree<BTEntry> tree) {
    ArrayList<BTEntry> entriesInOrder = new ArrayList<>();
    collectElementsInOrder(tree, entriesInOrder);
    StringBuilder keysInOrder = new StringBuilder();
    String comma = "";
    for (BTEntry bte : entriesInOrder) {
      keysInOrder.append(comma);
      keysInOrder.append(bte.key);
      comma = ",";
    }
    return keysInOrder.toString();
  }


  static boolean systemTest() {
    boolean success = true;
    BinTree<BTEntry> tree = new BinTree<>(new BTEntry("Aeneam", "bla"));
    ArrayList<BTEntry> inserts = new ArrayList<>();
    inserts.add(new BTEntry("Lorem", "bla"));
    inserts.add(new BTEntry("ipsum", "bla"));
    inserts.add(new BTEntry("dolor", "bla"));
    inserts.add(new BTEntry("sit", "bla"));
    inserts.add(new BTEntry("amet", "bla"));
    inserts.add(new BTEntry("consectetuer", "bla"));
    inserts.add(new BTEntry("adipiscing", "bla"));
    inserts.add(new BTEntry("elit", "bla"));

    for (BTEntry ins : inserts) {
      tree.insert(ins);
    }
    success &= searchTreeCondition(tree); //is it still a sorted tree?
    success &= isSorted(tree); //is it still a sorted tree?
    for (BTEntry ins : inserts) {
      success &= tree.contains(ins);
    }
    
    ArrayList<BTEntry> deletes = new ArrayList<>();
    deletes.add(new BTEntry("amet", ""));
    deletes.add(new BTEntry("elit", ""));
    deletes.add(new BTEntry("adipiscing", ""));
    deletes.add(new BTEntry("somethingElse", ""));

    for (BTEntry d : deletes) {
      tree.delete(d);
    }
    for (BTEntry d : deletes) {
      success &= !tree.contains(d); //were the elements really deleted?
    }
    success &= searchTreeCondition(tree); //is it still a sorted tree?
    success &= isSorted(tree); //is it still a sorted tree?

    return success;
  }

  static boolean searchTreeCondition(BinTree<BTEntry> tree) {
    Stack<BinTree<BTEntry>> stack = new Stack<>();
    BinTree<BTEntry> currentSubTree = null;
    stack.push(tree);
    int i = 0;
    while (!stack.isEmpty() && i++ < 4) {
      currentSubTree = stack.pop();
      
      if (tree.left != null && tree.right != null) {
        if (tree.left.content.compareTo(tree.content) >= 0)
          return false;
        if (tree.content.compareTo(tree.right.content) >= 0)
          return false;
      }
      
      if (tree.left != null) {
        stack.push(tree.left);
      }
      if (tree.right != null) {
        stack.push(tree.right);
      }
    }
    return true;
  }


  static boolean isSorted(BinTree<BTEntry> tree) {
    ArrayList<BTEntry> entriesInOrder = new ArrayList<>();
    collectElementsInOrder(tree, entriesInOrder);
    for (int i = 0; i < entriesInOrder.size() - 2; i++) {
      if (entriesInOrder.get(i).compareTo(entriesInOrder.get(i+1)) > 0)
        return false;
    }
    return true;
  }

  static void collectElementsInOrder(BinTree<BTEntry> tree, ArrayList<BTEntry> elements) {
    if (tree == null)
      return;
    collectElementsInOrder(tree.left, elements);
    elements.add(tree.content);
    collectElementsInOrder(tree.right, elements);
  }
  
  static class BTEntry implements Comparable<BTEntry> {
    final String key;
    final String value;
    BTEntry(String key, String value) {
      this.key = key;
      this.value = value;
    }

    public int compareTo(BTEntry other) {
      return key.compareTo(other.key);
    }
  } 
  
  static class BTInsertTest {
    BinTree<BTEntry> treeBefore;
    BTEntry element;
    String treeAfterAsString;
    BTInsertTest(BinTree<BTEntry> treeBefore, BTEntry element, String treeAfterAsString) {
      this.treeBefore = treeBefore;
      this.element = element;
      this.treeAfterAsString = treeAfterAsString;
    }
  }
}
